'''
Reflection functions.
'''

import types

def _public_attrs(obj, pred=lambda attr: True):
    '''Public attributes of an object, fields and methods,
    filtered by a predicate.
    '''
    return sorted([
        attr for attr in dir(obj)
        if not attr.startswith('_') and pred(attr)
    ])


def _is_instance_method(obj, attr):
    attribute = getattr(obj, attr)
    return (
        isinstance(attribute, types.MethodType) and
        attribute.im_class is obj.__class__
    )


def _is_instance_field(obj, attr):
    attribute = getattr(obj, attr, None)
    return (
        not isinstance(attribute, types.MethodType) and
        getattr(obj.__class__, attr, None) is None
    )


def _is_instance_property(obj, attr):
    attribute = getattr(obj, attr, None)

    return (
        isinstance(getattr(obj.__class__, attr, None), property)
    )


def class_attrs(obj):
    '''Public attributes(fields and methods) of an object's class.
    '''
    def pred(attr):
        return (
            (_is_instance_field(obj.__class__, attr) or
             _is_instance_method(obj.__class__, attr)) and
            not isinstance(getattr(obj.__class__, attr, None), property)
        )

    return _public_attrs(obj.__class__, pred)


def class_fields(obj):
    '''Public fields of an object's class.
    '''
    def pred(attr):
        return (
            _is_instance_field(obj.__class__, attr) and
            not isinstance(getattr(obj.__class__, attr, None), property)
        )
    return _public_attrs(obj.__class__, pred)


def class_methods(obj):
    '''Public methods on an object's class.
    '''
    def pred(attr):
        return (
            _is_instance_method(obj.__class__, attr) and
            not isinstance(getattr(obj.__class__, attr, None), property)
        )
    return _public_attrs(obj.__class__, pred)


def instance_attrs(obj):
    '''Public attributes(fields, properties, and methods) of an object.
    '''
    def pred(attr):
        return (
            _is_instance_field(obj, attr) or
            _is_instance_method(obj, attr) or
            _is_instance_property(obj, attr)
        )

    return _public_attrs(obj, pred)


def instance_members(obj):
    '''Public members (fields and properties) of an object.
    '''
    def pred(attr):
        return (
            _is_instance_field(obj, attr) or
            _is_instance_property(obj, attr)
        )

    return _public_attrs(obj, pred)


def instance_fields(obj):
    '''Public fields of an object.
    '''
    return _public_attrs(obj, lambda attr: _is_instance_field(obj, attr))


def instance_methods(obj):
    '''Public methods on an object.
    '''
    return _public_attrs(obj, lambda attr: _is_instance_method(obj, attr))


def instance_properties(obj):
    '''Public properties of an object.
    '''
    return _public_attrs(obj, lambda attr: _is_instance_property(obj, attr))



import random
import re
import collections

# ----------------------------------------------------------------------------
# Base Generators
# ----------------------------------------------------------------------------


def forever(initial=0, step=1):
    x = initial
    while True:
        yield x
        x += step


def build(factory):
    while True:
        yield factory()


def duplicate(anything):
    while True:
        yield anything


def zeros():
    while True:
        yield 0


def ones():
    while True:
        yield 1


def random_ints(start=0, end=1):
    while True:
        yield random.randint(start, end)


def random_floats(start=0.0, end=1.0):
    while True:
        yield random.uniform(start, end)


# ----------------------------------------------------------------------------
# Take
# ----------------------------------------------------------------------------


def take(n_or_predicate, iterable):
    if callable(n_or_predicate):
        for x in take_while(n_or_predicate, iterable):
            yield x

    else:
        for x in take_n(n_or_predicate, iterable):
            yield x


def take_n(n, iterable):
    i = 0
    for x in iterable:
        i += 1
        if i > n:
            break
        yield x


def take_while(predicate, iterable):
    for x in iterable:
        if predicate(x):
            yield x
        else:
            break

# ----------------------------------------------------------------------------
# Drop
# ----------------------------------------------------------------------------


def drop(n_or_predicate, iterable):
    if callable(n_or_predicate):
        for x in drop_while(n_or_predicate, iterable):
            yield x

    else:
        for x in drop_n(n_or_predicate, iterable):
            yield x


def drop_n(n, iterable):
    i = 0
    for x in iterable:
        i += 1
        if i <= n:
            continue

        yield x


def drop_while(predicate, iterable):
    for x in iterable:
        if predicate(x):
            continue
        else:
            yield x


# ----------------------------------------------------------------------------
# Value Selectors
# ----------------------------------------------------------------------------


def nth(n, iterable):
    v = None
    for x in take_n(n, iterable):
        v = x

    return v


def head(iterable):
    for x in iterable:
        return x


def tail(iterable):
    skipped = False
    for x in iterable:
        if not skipped:
            skipped = True
            continue
        else:
            yield x


# ----------------------------------------------------------------------------
# Indexing
# ----------------------------------------------------------------------------


def index(needle, iterable):
    for i, x in enumerate(iterable):
        if x == needle:
            return i

    raise IndexError(needle)


def indexes(needle, iterable):
    for i, x in enumerate(iterable):
        if x == needle:
            yield i


# ----------------------------------------------------------------------------
# Filtering
# ----------------------------------------------------------------------------

def select(regex_or_needle_or_predicate, iterable):
    if callable(regex_or_needle_or_predicate):
        for x in select_where(regex_or_needle_or_predicate, iterable):
            yield x

    else:
        try:
            for x in select_match(regex_or_needle_or_predicate, iterable):
                yield x
            return

        except TypeError as err:
            if 'first argument' not in str(err):
                raise err
            else:
                pass

        for x in select_eq(regex_or_needle_or_predicate, iterable):
            yield x


def select_eq(needle, iterable):
    for x in iterable:
        if x == needle:
            yield x


def select_where(predicate, iterable):
    for x in iterable:
        if predicate(x):
            yield x


def select_match(regex, iterable):
    compiled = re.compile(regex)
    for x in iterable:
        if compiled.match(x) is not None:
            yield x


def reject(regex_or_needle_or_predicate, iterable):
    if callable(regex_or_needle_or_predicate):
        for x in reject_where(regex_or_needle_or_predicate, iterable):
            yield x

    else:
        try:
            for x in reject_match(regex_or_needle_or_predicate, iterable):
                yield x
            return

        except TypeError as err:
            if 'first argument' not in str(err):
                raise err
            else:
                pass

        for x in reject_eq(regex_or_needle_or_predicate, iterable):
            yield x


def reject_eq(needle, iterable):
    for x in iterable:
        if x != needle:
            yield x


def reject_where(predicate, iterable):
    for x in iterable:
        if not predicate(x):
            yield x


def reject_match(regex, iterable):
    compiled = re.compile(regex)
    for x in iterable:
        if compiled.match(x) is None:
            yield x


def partition(needle_or_predicate, iterable):
    if callable(needle_or_predicate):
        return partition_where(needle_or_predicate, iterable)

    else:
        try:
            return partition_match(needle_or_predicate, iterable)
        except TypeError as err:
            if 'first argument' not in str(err):
                raise err
            else:
                pass

        return partition_eq(needle_or_predicate, iterable)


def partition_eq(needle, iterable):
    left = collections.deque()
    right = collections.deque()
    iterator = iter(iterable)

    def selector():
        while True:
            if len(left) > 0:
                yield left.popleft()
            else:
                try:
                    item = iterator.next()
                    if item == needle:
                        yield item
                    else:
                        right.append(item)

                except StopIteration:
                    return

    def rejector():
        while True:
            if len(right) > 0:
                yield right.popleft()
            else:
                try:
                    item = iterator.next()
                    if item != needle:
                        yield item
                    else:
                        left.append(item)

                except StopIteration:
                    return

    return selector(), rejector()


def partition_where(predicate, iterable):
    left = collections.deque()
    right = collections.deque()
    iterator = iter(iterable)

    def selector():
        while True:
            if len(left) > 0:
                yield left.popleft()
            else:
                try:
                    item = iterator.next()
                    if predicate(item):
                        yield item
                    else:
                        right.append(item)

                except StopIteration:
                    return

    def rejector():
        while True:
            if len(right) > 0:
                yield right.popleft()
            else:
                try:
                    item = iterator.next()
                    if not predicate(item):
                        yield item
                    else:
                        left.append(item)

                except StopIteration:
                    return

    return selector(), rejector()


def partition_match(regex, iterable):
    compiled = re.compile(regex)
    left = collections.deque()
    right = collections.deque()
    iterator = iter(iterable)

    def selector():
        while True:
            if len(left) > 0:
                yield left.popleft()
            else:
                try:
                    item = iterator.next()
                    if compiled.match(item) is not None:
                        yield item
                    else:
                        right.append(item)

                except StopIteration:
                    return

    def rejector():
        while True:
            if len(right) > 0:
                yield right.popleft()
            else:
                try:
                    item = iterator.next()
                    if compiled.match(item) is None:
                        yield item
                    else:
                        left.append(item)

                except StopIteration:
                    return

    return selector(), rejector()



import copy


def updated(left, right, join=lambda l, r: r):
    result = copy.copy(left)

    for k, v in right.iteritems():
        try:
            result[k] = join(left[k], v)
        except KeyError:
            result[k] = v

    return result



from functools import wraps


def lazy(func):
    cache = {}

    @wraps(func)
    def wrap(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key not in cache:
            cache[key] = func(*args, **kwargs)

        return cache[key]

    return wrap


def lazy_property(func):
    canary = object()

    prop_name = '_prop_' + func.func_name

    @property
    @wraps(func)
    def wrap(self, *args, **kwargs):
        value = getattr(self, prop_name, canary)
        if value is canary:
            value = func(self, *args, **kwargs)
            setattr(self, prop_name, value)

        return value

    return wrap



from reflection import class_attrs, class_fields, class_methods
from reflection import instance_attrs, instance_members, instance_fields, instance_properties, instance_methods

from decorators import lazy, lazy_property

from collection import updated

from generators import forever, build, duplicate, zeros, ones
from generators import random_ints, random_floats
from generators import take, take_n, take_while
from generators import drop, drop_n, drop_while
from generators import nth, head, tail
from generators import select, select_where, select_eq, select_match
from generators import reject, reject_where, reject_eq, reject_match
from generators import partition, partition_where, partition_eq, partition_match


